/* 
Version 2014-12-15
 Realisiert mit Tiny25
 
RC mit 8MHz, 

 Layout:
                     ADC0   Reset  1    8  +5V
                Endschalter   PB3  2     7  PB2 Poti "Speed", ADC1, Endschalter
                              PB4  3     6  PB1 Step
                              GND  4     5  PB0 Dir   und LED
  
*/

/* Optimizer auf -Os */

#define F_CPU 8000000UL	/* CPU clock in Hertz */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <util/atomic.h>

#define U8 unsigned char
#define U16 unsigned int
#define I16 signed int
#define U32 unsigned long
#define I32 signed long

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128


typedef struct{
	I16 pos;
	I16 endpos;
	I16 speed;
//	I16 angle;
	I32 vHiAccFw;
	I32 vLoAccFw;
	I32 vHiAccRev;
	I32 vLoAccRev;
	U8  fw;
}Tglob;

Tglob volatile g;

#define STEP_BIT      BIT1
#define DIR_BIT       BIT0
#define ENDSWITCH_BIT BIT3

#define DIR_POS       PORTB  |=		DIR_BIT
#define DIR_NEG       PORTB  &=~   	DIR_BIT
#define SET_STEP   	  PORTB  |= 	STEP_BIT
#define CLR_STEP   	  PORTB  &=~	STEP_BIT
#define END_SWITCH 	  (PINB   &     ENDSWITCH_BIT)


//------------------------------------------------------------
#define Vmax 	1000000L  //1.000.000
#define acc 	500
#define angle 	2950 //3200 = 360
//------------------------------------------------------------

#define rampSteps (Vmax/acc/2)



//-----------------------------------------------------------------------------------------------
ISR(TIM0_COMPA_vect){ // 83,3KHz  Fhrt Motor mit g.v, g.acc
//-----------------------------------------------------------------------------------------------
 static I32 v=0;
 static I32 sum=0;
	CLR_STEP;
	
	if (g.fw){
		if ( v < g.vHiAccFw){
		  v += acc;
		}else
	    if ( v > g.vLoAccFw){
		  v -= acc;
		}
    }else{
		if ( v < g.vHiAccRev){
		  v += acc;
		}else
	    if ( v > g.vLoAccRev){
		  v -= acc;
		}
	} 
	//Ist der nchste Schritt fllig?
	sum += v;
	if (sum >= Vmax){
		//make a pos step
		sum -= Vmax;
		DIR_POS;
		SET_STEP; // do the step
		g.pos++;

		if (g.pos >= g.endpos){ //Rckfahrt starten
			g.fw=0;
		}
	}else
	if (sum <= -Vmax){
		//make a neg step
		sum += Vmax;
		DIR_NEG;
		SET_STEP; // do the step
		g.pos--;
		if (g.pos < 0){  //Hinfahrt starten (rampSteps werden bis zum Stillstand der Rckfahrt gebraucht)
			g.fw=1;
		}
  	}
}

//-----------------------------------------------------------------------------------------------
int main(void){
//-----------------------------------------------------------------------------------------------
I32 vFwd;
I16 i=100;
I32 vHiAccFw;
I32 vLoAccFw;

	cli();
	DDRB = BIT0 | BIT1;
	PORTB = ENDSWITCH_BIT; //Pull-Up fr Endschalter

	ACSR=(1<<ACD); //Analogkomparator aus
	PRR=(1<<PRUSI); //USI aus

	//Timer 0: 10KHz Interrupts 
	TCCR0A=(1<<WGM01);  //CTC
	TCCR0B=(1<<CS01);   //Teiler 8 (1MHz)

	OCR0A = 11; // Teiler / 12 -> 83,3KHz -> 1562 1/min
	OCR0B = 0; 

	GTCCR = 0;
	TIMSK =(1<<OCIE0A);

	ADMUX = 1;  //REFS0=0, REFS1=0, ADLAR=0:  Vcc als Referenz
	ADCSRB = 0;

	g.pos  = 0 - rampSteps; //Motor luft von - rampSteps bis angle - rampSteps
							//Damit ist der Start der Bremsrampe genau bei 0.
							//Das ist in Assembler mit weniger Code detektierbar
	
	_delay_ms(100);
	ADCSRA = (1<<ADEN) | (1<<ADSC) | (1<<ADPS2) | (1<<ADPS1); //Teiler /64  -> 125KHz
	//Warte bis Messung fertig.
	while ( ADCSRA & (1<<ADSC))
	;
	g.speed = ADC+1;
	vFwd = Vmax / 32768L * g.speed;//Langsam vorwrts auf Endschalter zufahren
    g.endpos = angle - rampSteps;
	g.fw=1;
	g.vHiAccRev = 0 - Vmax + acc;
	g.vLoAccRev = 0 - Vmax - acc;
	vFwd = Vmax / 32768L * g.speed;
	g.vHiAccFw = vFwd + acc;   
	g.vLoAccFw = vFwd - acc;
	sei(); //Interrupts scharf.

//...................................................................................... 
	while(1){
		if (END_SWITCH){
			ATOMIC_BLOCK(ATOMIC_FORCEON){
				g.pos = g.endpos;
			}
		}
	    if (g.fw && (g.pos < 2000)){  //Nur wenn Verzgerung kein Problem ist.
				_delay_us(100);
			if (--i < 0){ //Gelegentlich messen. 
				i = 300;	
				ADCSRA = (1<<ADEN) | (1<<ADSC) | (1<<ADPS2) | (1<<ADPS1); //Teiler /64  -> 125KHz
				while ( ADCSRA & (1<<ADSC)) //Warte bis Messung fertig.
					;
				g.speed = ADC+1;
				vFwd = Vmax / 32768L * g.speed;
				vHiAccFw = vFwd + acc;
				vLoAccFw = vFwd - acc;
				
			
				ATOMIC_BLOCK(ATOMIC_FORCEON){  //<--- In .lss prfen, ob cli/sei nur die Zuweisungen kapselt.  
					g.vHiAccFw = vHiAccFw;         //     Der Optimizer verschiebt das sei() manchmal.
				}

				ATOMIC_BLOCK(ATOMIC_FORCEON){
					g.vLoAccFw = vLoAccFw;
				}
				
				/* Beispielcode fr korrekte ATOMIC Kapselung:

						 30e:	f8 94       	cli
						 310:	60 93 6e 00 	sts	0x006E, r22
						 314:	70 93 6f 00 	sts	0x006F, r23
						 318:	80 93 70 00 	sts	0x0070, r24
						 31c:	90 93 71 00 	sts	0x0071, r25
						 320:	78 94       	sei

					Leider macht der olle Optimizer daraus gerne sowas:
						 362:	f8 94       	cli
						 364:	6c 50       	subi	r22, 0x0C	; 12
						 366:	7e 4f       	sbci	r23, 0xFE	; 254
						 368:	8f 4f       	sbci	r24, 0xFF	; 255
						 36a:	9f 4f       	sbci	r25, 0xFF	; 255
						 36c:	60 93 6e 00 	sts	0x006E, r22
						 370:	70 93 6f 00 	sts	0x006F, r23
						 374:	80 93 70 00 	sts	0x0070, r24
						 378:	90 93 71 00 	sts	0x0071, r25
						 37c:	78 94       	sei




				*/
			}
		}
	}
}
//EOF
